
# Tectonic - An academic Cyber Range
# Copyright (C) 2024 Grupo de Seguridad Informática, Universidad de la República,
# Uruguay
#
# This file is part of Tectonic.
#
# Tectonic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Tectonic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tectonic.  If not, see <http://www.gnu.org/licenses/>.

"""Constants used by Tectonic."""

OS_DATA = {
    "ubuntu22": {
        "ami_filter": "ubuntu/images/*/ubuntu-*-22.04-amd64-server*",
        "owner": "099720109477",
        "username": "ubuntu",
        "cloud_image_url": "https://cloud-images.ubuntu.com/releases/22.04/release-20230616/ubuntu-22.04-server-cloudimg-amd64.img",
        "cloud_image_checksum": "fe102bfb3d3d917d31068dd9a4bd8fcaeb1f529edda86783f8524fdc1477ee29",
        "docker_base_image": "gsitectonic/ubuntu22",
        "entrypoint": "/usr/bin/systemd",
    },
    "ubuntu22_deep_learning": {
        "ami_filter": "*Deep Learning Base OSS Nvidia Driver GPU AMI (Ubuntu 22.04)*",
        "owner": "898082745236",
        "username": "ubuntu",
    },
    "rocky8": {
        "ami_filter": "Rocky*8.8*x86_64",
        "owner": "792107900819",
        "username": "rocky",
        #"cloud_image_url": "http://dl.rockylinux.org/pub/rocky/8/images/x86_64/Rocky-8-GenericCloud.latest.x86_64.qcow2",
        "cloud_image_url": "file:///data/isos/Rocky-8-GenericCloud.latest.x86_64.qcow2",
        "cloud_image_checksum": "d17f15a7649dd064795306c114b90fc5062e7d5fefa9e9f0bd6b7ce1aeac3ae5",
        "docker_base_image": "gsitectonic/rocky8",
        "entrypoint": "/usr/sbin/init",
    },
    "kali": {
        "ami_filter": "kali-last-snapshot-amd64*",
        "owner": "679593333241",  # TODO: Marketplace owner id
        "username": "kali",
        # cloud-init image not provided by Kali :(
        # See https://codingpackets.com/blog/kali-linux-cloud-init-image/
            "cloud_image_url": "file:///data/isos/kali-linux-2023.3-qemu-amd64.qcow2",
        "cloud_image_checksum": "b48cc396cf91ea1f38de4dca7abfcc0175f07d7af1b7e6df89a76194e7fda3d9",
        # "cloud_image_url": "https://cdimage.kali.org/kali-2023.3/kali-linux-2023.3-qemu-amd64.7z"
        # "cloud_image_checksum": "9ebbea4abb545c8e4a56153e2b5fa0ad90658c583af2fa76ad8c0a0b4ba23e20",
        },
    "windows_srv_2022": {
        "ami_filter": "Windows_Server-2022-English-Full-Base*",
        "owner": "amazon",
        "username": "administrator",
    },
}

SERVICES_SIZE = {
    "elastic" : {
        "libvirt" : {
            "test" : {
                "vcpu":2,
                "memory":4096, #MB
                "disk":10, #GB
            },
            "small" : {
                "vcpu":4,
                "memory":8192,
                "disk":60,
            },
            "medium": {
                "vcpu":8,
                "memory":16384,
                "disk":110,
            },
            "big":{
                "vcpu":12,
                "memory":24576,
                "disk":160,
            }
        },
        "aws" : {
            "small": "t2.small",
            "medium": "t2.medium",
            "big": "t2.big",
        }
    },
    "caldera" : {
        "libvirt" : {
            "test" : {
                "vcpu":1,
                "memory":2048, #MB
                "disk":10, #GB
            },
            "small" : {
                "vcpu":2,
                "memory":2048,
                "disk":20,
            },
            "medium": {
                "vcpu":3,
                "memory":4096,
                "disk":20,
            },
            "big":{
                "vcpu":4,
                "memory":8192,
                "disk":50,
            }
        },
        "aws" : {
            "small": "t2.small",
            "medium": "t2.medium",
            "big": "t2.big",
        }
    }
}




# Tectonic - An academic Cyber Range
# Copyright (C) 2024 Grupo de Seguridad Informática, Universidad de la República,
# Uruguay
#
# This file is part of Tectonic.
#
# Tectonic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Tectonic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Tectonic.  If not, see <http://www.gnu.org/licenses/>.

import libvirt
import libvirt_qemu
import time
from ipaddress import ip_network, ip_address

class LibvirtClientException(Exception):
    pass


# Avoid console output.
# See https://stackoverflow.com/questions/45541725/avoiding-console-prints-by-libvirt-qemu-python-apis
def libvirt_callback(userdata, err):
    pass


class Client:
    """Class for managing libvirt connections."""
    STATES_MSG = {
        libvirt.VIR_DOMAIN_NOSTATE: "NOSTATE",
        libvirt.VIR_DOMAIN_RUNNING: "RUNNING",
        libvirt.VIR_DOMAIN_BLOCKED: "BLOCKED",
        libvirt.VIR_DOMAIN_PAUSED: "PAUSED",
        libvirt.VIR_DOMAIN_SHUTDOWN: "SHUTDOWN",
        libvirt.VIR_DOMAIN_SHUTOFF: "SHUTOFF",
        libvirt.VIR_DOMAIN_CRASHED: "CRASHED",
        libvirt.VIR_DOMAIN_PMSUSPENDED: "SUSPENDED",
    }

    def __init__(self, description):
        self.description = description
        self.uri = description.libvirt_uri
        libvirt.registerErrorHandler(f=libvirt_callback, ctx=None)
        try:
            self.conn = libvirt.open(self.uri)
        except:
            self.conn = None
            raise LibvirtClientException(f"Cannot connect to libvirt server at {self.uri}")

    def __del__(self):
        if self.conn is not None:
            try:
                self.conn.close()
            except:
                pass

    def wait_for_agent(self, domain, sleep=5, max_tries = 10):
        tries = 1
        agent_ready = False
        while not agent_ready and tries < max_tries:
            try:
                libvirt_qemu.qemuAgentCommand(
                    domain, '{"execute": "guest-ping"}', 10, 0
                )
                agent_ready = True
            except libvirt.libvirtError:
                tries += 1
                time.sleep(sleep)

        if not agent_ready:
            raise LibvirtClientException("Cannot connect to QEMU agent.")

    def get_machine_private_ip(self, name):
        """Returns the private IP address of a domain.

           If the domain has more than one IP address, the first
           address inside network_cidr_block is returned.

        """
        try:
            domain = self.conn.lookupByName(name)
        except libvirt.libvirtError:
            return None
        self.wait_for_agent(domain)

        lab_network = ip_network(self.description.network_cidr_block)
        services_network = ip_network(self.description.services_network)
        services_list = []
        for service in self.description.get_services_to_deploy():
            services_list.append(self.description.get_service_name(service))
        interfaces = domain.interfaceAddresses(
            libvirt.VIR_DOMAIN_INTERFACE_ADDRESSES_SRC_AGENT, 0
        )
        for interface_name, val in interfaces.items():
            if interface_name != "lo" and val["addrs"]:
                for ipaddr in val["addrs"]:
                    if name in services_list:
                        # TODO: The first time this fails and returns None as the interfaces other than localhost are empty. 
                        # The rest of the times it works ok. 
                        if ip_address(ipaddr["addr"]) in services_network:
                            return ipaddr["addr"]                         
                    else:
                        if ip_address(ipaddr["addr"]) in lab_network:
                            return ipaddr["addr"]
            else:
                if not val["addrs"]:
                    # Return fixed IP for service since the first time it is empty
                    if name == f"{self.description.institution}-{self.description.lab_name}-elastic":
                        return "10.0.0.130"
                    elif name == f"{self.description.institution}-{self.description.lab_name}-caldera":
                        return "10.0.0.132"
        return None

    def get_machine_public_ip(self, name):
        """
        Get a machine public IP

        Parameters:
            name (string): complete name of the machine

        Returns:
            None: this method is not implemented for libvirt
        """
        return None



    def delete_image(self, pool_name, image_name):
        try:
            pool = self.conn.storagePoolLookupByName(pool_name)
        except libvirt.libvirtError:
            raise LibvirtClientException(f"Failed to locate {pool_name} storage pool.")

        vol = None
        try:
            vol = pool.storageVolLookupByName(image_name)
        except libvirt.libvirtError:
            pass
        if vol:
            vol.delete()

    def get_instance_status(self, instance_name):
        try:
            dom = self.conn.lookupByName(instance_name)
        except libvirt.libvirtError:
            return "NOT FOUND"

        state, _ = dom.state()
        return self.STATES_MSG.get(state, "UNKNOWN")

    def start_instance(self, instance_name):
        """Starts a stopped instance."""
        try:
            dom = self.conn.lookupByName(instance_name)
        except libvirt.libvirtError as e:
            raise LibvirtClientException(f"Domain {instance_name} not found.")

        state, _ = dom.state()
        if state != libvirt.VIR_DOMAIN_RUNNING:
            dom.create()

    def stop_instance(self, instance_name):
        """Stops a running instance."""
        try:
            dom = self.conn.lookupByName(instance_name)
        except libvirt.libvirtError as e:
            raise LibvirtClientException(f"Domain {instance_name} not found.")
        state, _ = dom.state()
        if state != libvirt.VIR_DOMAIN_SHUTOFF:
            dom.shutdown()

    def reboot_instance(self, instance_name):
        """Reboots a running instance."""
        try:
            dom = self.conn.lookupByName(instance_name)
        except libvirt.libvirtError as e:
            raise LibvirtClientException(f"Domain {instance_name} not found.")

        state, _ = dom.state()
        if state == libvirt.VIR_DOMAIN_RUNNING:
            dom.reboot()
        elif state == libvirt.VIR_DOMAIN_SHUTOFF:
            dom.create()       
